"""
Write languoids data to csv files
"""
from pyglottolog.cli_util import add_output_dir


def register(parser):
    add_output_dir(parser)


def run(args):
    res = args.repos.write_languoids_table(args.output, args.repos_version)
    print('Lagnuoids table written to')
    for p in res:
        print('{0}'.format(p))



import pathlib

from clldutils.clilib import ParserError, add_format, PathType

__all__ = ['add_output_dir']


def add_output_dir(parser):
    parser.add_argument(
        '--output',
        help='An existing directory for the output',
        type=PathType(type='dir'),
        default=pathlib.Path('.'),
    )


def register_search(parser, example):
    parser.add_argument(
        'query', metavar="QUERY", help='Search query, e.g. "{0}"'.format(example))
    add_format(parser, default='simple')


def get_languoid(args, spec):
    if spec:
        lang = args.repos.languoid(spec)
        if not lang:
            raise ParserError('Invalid languoid {0}'.format(spec))
        return lang



import itertools
import collections

from clldutils.path import TemporaryDirectory
from pycldf.dataset import Dataset
import cldfzenodo


def read_grouped_cldf_languages(doi):  # pragma: no cover
    rec = cldfzenodo.Record.from_doi(doi)
    with TemporaryDirectory() as tmp:
        ds = Dataset.from_metadata(rec.download_dataset(tmp))
        langs = sorted(
            ds.iter_rows('LanguageTable', 'id', 'glottocode', 'name'),
            key=lambda r: r['glottocode'] or '')

    for gc, langs in itertools.groupby(langs, lambda d: d['glottocode']):
        yield gc, list(langs)


class LinkProvider(object):
    __domain__ = None
    __doi__ = None
    __url_template__ = None
    __label_template__ = None

    def __init__(self, repos=None):
        self.repos = repos

    def iterupdated(self, languoids):  # pragma: no cover
        if self.__domain__ and self.__doi__ and self.__url_template__:
            # FIXME: Ideally, we'd want the CLDF data to specify full URLs for languages via a
            # valueUrl property on the ID column (or the LanguageTable?).
            res = collections.defaultdict(list)
            for gc, langs in read_grouped_cldf_languages(self.__doi__):
                for lang in langs:
                    lang = {k: v.strip() if isinstance(v, str) else v for k, v in lang.items()}
                    item = self.__url_template__.format(lang)
                    if self.__label_template__:
                        item = (item, self.__label_template__.format(lang))
                    res[gc].append(item)
            for lang in languoids:
                if lang.update_links(self.__domain__, res.get(lang.id, [])):
                    yield lang
        else:
            raise NotImplementedError()


class PHOIBLE(LinkProvider):  # pragma: no cover
    __domain__ = 'phoible.org'
    __doi__ = '10.5281/zenodo.2562766'

    def iterupdated(self, languoids):  # pragma: no cover
        urls = {}
        for gc, langs in read_grouped_cldf_languages(self.__doi__):
            urls[gc] = ['https://{0}/languages/{1}'.format(self.__domain__, gc)]
        for lang in languoids:
            if lang.update_links(self.__domain__, urls.get(lang.id, [])):
                yield lang


class APICS(LinkProvider):  # pragma: no cover
    __domain__ = "apics-online.info"
    __doi__ = '10.5281/zenodo.3823887'
    __url_template__ = 'https://' + __domain__ + '/contributions/{0[id]}'
    __label_template__ = '{0[name]}'


class WALS(LinkProvider):  # pragma: no cover
    __domain__ = "wals.info"
    __doi__ = '10.5281/zenodo.3606197'
    __url_template__ = 'https://' + __domain__ + '/languoid/lect/wals_code_{0[id]}'
    __label_template__ = '{0[name]}'



"""
Create an HTML/Javascript map (using leaflet) of Glottolog languoids.

Languoids to plot can be specified as glottocodes or comma-separated
"glottocode,latitude,longitude" triples in a file or as stdin.
"""
import sys
import json
import string
import webbrowser
import collections

from clldutils.color import qualitative_colors
from clldutils.path import git_describe
from clldutils.clilib import PathType

from pyglottolog.cli_util import add_output_dir


def register(parser):
    add_output_dir(parser)
    parser.add_argument(
        '--glottocodes',
        help='File listing Glottocodes the map should be restricted to',
        type=PathType(type='file'),
        default=None,
    )
    parser.add_argument(
        '--min-langs-for-legend',
        help='minimum number of languages in a family for a map legend entry',
        type=int,
        default=10,
    )
    parser.add_argument(
        '--open',
        action='store_true',
        default=False,
        help='open the map in a web browser after',
    )


def get_glottocodes(args):
    glottocodes = []
    if args.glottocodes:
        glottocodes = args.glottocodes.read_text(encoding='utf8').split()
    elif not args.test:  # pragma: no cover
        glottocodes = [line.strip() for line in sys.stdin.readlines()]
    res = {}
    for gc in glottocodes:
        if ',' in gc:
            gc, lat, lon = gc.split(',')
            res[gc] = (float(lat) if lat else None, float(lon) if lon else None)
        else:
            res[gc] = (None, None)
    return res


def run(args):
    legend = collections.Counter()
    glottocodes = get_glottocodes(args)

    nodes = {n.id: n for n in args.repos.languoids()}
    langs = []
    for n in nodes.values():
        if ((not glottocodes and n.level == args.repos.languoid_levels.language)
                or (glottocodes and n.id in glottocodes)) \
                and (n.latitude != None or glottocodes.get(n.id, (None,))[0] != None):  # noqa: W503
            fid = n.lineage[0][1] if n.lineage else n.id
            if (not nodes[fid].category.startswith('Pseudo')) or fid == n.id:
                langs.append((n, fid))
                legend.update([fid])

    color_map = [fid for fid, _ in legend.most_common()]
    color_map = dict(zip(color_map, qualitative_colors(len(color_map))))

    def l2f(t):
        n, fid = t
        lat, lon = glottocodes.get(n.id, (None, None))
        if lat is None:
            lat = n.latitude
        if lon is None:
            lon = n.longitude
        if lon <= -26:
            lon += 360  # make the map pacific-centered.

        return {
            "geometry": {"coordinates": [lon, lat], "type": "Point"},
            "id": n.id,
            "properties": {
                "name": n.name,
                "color": color_map[fid],
                "family": nodes[fid].name,
                "family_id": fid,
            },
            "type": "Feature"
        }

    def legend_item(fid, c):
        return \
            '<span style="background-color: {0}; border: 1px solid black;">'\
            '&nbsp;&nbsp;&nbsp;</span> '\
            '<a href="https://glottolog.org/resource/languoid/id/{1}">{2}</a> ({3})'.format(
                color_map[fid], fid, nodes[fid].name, c)

    geojson = {
        "features": list(map(l2f, langs)),
        "properties": {
            "legend": {
                fid: legend_item(fid, c) for fid, c in legend.most_common() if
                c >= args.min_langs_for_legend},
        },
        "type": "FeatureCollection"
    }

    def rendered_template(name, **kw):
        return string.Template(
            args.pkg_dir.joinpath('templates', 'htmlmap', name).read_text(encoding='utf8')
        ).substitute(**kw)

    jsname = 'glottolog_map.json'
    args.output.joinpath(jsname).write_text(
        rendered_template('htmlmap.js', geojson=json.dumps(geojson, indent=4)), encoding='utf8')
    html = args.output.joinpath('glottolog_map.html')
    html.write_text(
        rendered_template(
            'htmlmap.html',
            version=git_describe(args.repos.repos),
            jsname=jsname,
            nlangs=len(langs)),
        encoding='utf8')
    print(html.resolve().as_uri())
    if args.open:
        webbrowser.open(str(html))  # pragma: no cover



"""
Update a refprovider bib from an updated source bib.
"""
from clldutils.clilib import PathType
from clldutils.path import TemporaryDirectory

from linglit.bibtex import merge


def register(parser):
    parser.add_argument(
        'bibkey',
        choices=['benjamins', 'glossa', 'langsci', 'dplace', 'ldh', 'jocp'],
    )
    parser.add_argument(
        'bib',
        type=PathType(type='file'),
        help='path to updated source bib')
    parser.add_argument(
        '--merge',
        action='store_true',
        default=False,
    )


def run(args):  # pragma: no cover
    with TemporaryDirectory() as tmp:
        if args.merge:
            merge(args.bib, tmp / args.bib.name, delatex=True)
            args.bib = tmp / args.bib.name
        args.repos.bibfiles[args.bibkey + '.bib'].update(args.bib, log=args.log)



# hhtypes.py

import re
import functools
import itertools

from clldutils.misc import lazyproperty

from ..util import Trigger
from ..config import get_ini

__all__ = ['HHType', 'HHTypes']


@functools.total_ordering
class HHType(object):

    def __init__(self, s, p):
        self.name = s
        self.id = p.get(s, 'id')
        self.rank = p.getint(s, 'rank')
        self.abbv = p.get(s, 'abbv')
        self.bibabbv = p.get(s, 'bibabbv')
        self.description = p.get(s, 'description')
        self.triggers = [Trigger('hhtype', self.id, t)
                         for t in p.get(s, 'triggers').strip().splitlines() or []]

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.id} rank={self.rank}>'

    def __eq__(self, other):
        return self.rank == other.rank

    def __lt__(self, other):
        return self.rank < other.rank


class HHTypes(object):

    _rekillparen = re.compile(r" \([^)]*\)")

    _respcomsemic = re.compile(r"[;,]\s?")

    def __init__(self, fpath):
        ini = get_ini(fpath, interpolation=None)
        self._types = sorted([HHType(s, ini) for s in ini.sections()], reverse=True)
        self._type_by_id = {t.id: t for t in self._types}

    @classmethod
    def parse(cls, s):
        return cls._respcomsemic.split(cls._rekillparen.sub("", s))

    def __iter__(self):
        return iter(self._types)

    def __len__(self):
        return len(self._types)

    def __contains__(self, item):
        return item in self._type_by_id

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._types[item]
        return self._type_by_id.get(item, self._type_by_id.get('unknown'))

    @lazyproperty
    def triggers(self):
        flattened = itertools.chain.from_iterable(t.triggers for t in self)
        return list(flattened)



"""
Write simple languoid stats to build/languoids.json.

This is to allow comparison between two branches of the repos.

Intended usage:
```
git checkout master
glottolog-admin writelanguoidstats
git checkout <OTHER_BRANCH>
glottolog-admin check --old-languoids
```
"""
try:
    from git import Repo
except ImportError:  # pragma: no cover
    Repo = None

from clldutils import jsonlib


def run(args):  # pragma: no cover
    if Repo:
        assert str(Repo(str(args.repos.repos)).active_branch) == 'master', \
            'Command should be run on master branch'
    res = {'language': [], 'family': [], 'dialect': []}
    for lang in args.repos.languoids():
        res[lang.level.name].append(lang.id)
    jsonlib.dump(res, args.repos.build_path('languoids.json'))



"""
Update metadata.json from config/publication.ini
"""
import json
import configparser

from clldutils.metadata import Metadata


def run(args):  # pragma: no cover
    cfg = configparser.ConfigParser()
    cfg.read(str(args.repos.path('config', 'publication.ini')))
    md = {'title' if k == 'name' else k: v for k, v in cfg['web'].items()}
    md['publisher.contact'] = cfg['web']['contact']
    for k in ['publisher', 'license']:
        md.update(**{k + '.' + kk: v for kk, v in cfg[k].items()})

    args.repos.path('metadata.json').write_text(
        json.dumps(Metadata.from_jsonld({}, defaults=md).to_jsonld(), indent=2))



"""
Index
- all bib files for use with `glottolog refsearch`
- all languoid info files for use with `glottolog langsearch`

This will take about
- about 15 minutes to create an index of about 450 MB for references and
- a couple of minutes and create an index of about 60 MB for languoids.
"""
from pyglottolog import fts


def run(args):
    fts.build_index(args.repos, args.log)
    fts.build_langs_index(args.repos, args.log)



"""
Query wikidata's SPARQL endpoint for info about Glottolog languoids identified by Glottocode.

Run

prefix schema: <http://schema.org/>
SELECT ?item ?glottocode ?wikipedia WHERE {
    ?item wdt:P1394 ?glottocode.
    OPTIONAL {
        ?wikipedia schema:about ?item .
        ?wikipedia schema:inLanguage "en" .
        FILTER (SUBSTR(str(?wikipedia), 1, 25) = "https://en.wikipedia.org/")
    }
}

at https://query.wikidata.org/ and download the results to
build/glottocode2wikidata.csv
"""
from csvw.dsv import reader
import requests

from .util import LinkProvider

SPARQL = """\
prefix schema: <http://schema.org/>
SELECT ?item ?glottocode ?wikipedia WHERE {
    ?item wdt:P1394 ?glottocode.
    OPTIONAL {
        ?wikipedia schema:about ?item .
        ?wikipedia schema:inLanguage "en" .
        FILTER (SUBSTR(str(?wikipedia), 1, 25) = "https://en.wikipedia.org/")
    }
}"""


class Wikidata(LinkProvider):
    def iterupdated(self, languoids):  # pragma: no cover
        res = requests.post(
            'https://query.wikidata.org/sparql',
            data=dict(query=SPARQL),
            headers=dict(Accept='text/csv')
        )
        res = {}
        if self.repos:
            res = {d['glottocode']: d for d in reader(
                self.repos.path('build', 'glottocode2wikidata.csv'), dicts=True)}
        assert res
        for lang in languoids:
            urls = {
                'www.wikidata.org': [
                    res[lang.id]['item'].replace('http:', 'https:')] if lang.id in res else [],
                'en.wikipedia.org':
                    [res[lang.id]['wikipedia']]
                    if (lang.id in res) and res[lang.id]['wikipedia'] else [],
            }
            if any([lang.update_links(d, u) for d, u in urls.items()]):
                # Note: We must use list comprehension rather than a generator as first argument
                # to `any` to make sure `update_links` is called for each item in urls!
                yield lang



"""
Search Glottolog references
"""
from clldutils.clilib import Table, ParserError

from pyglottolog.cli_util import register_search
from pyglottolog import fts


def register(parser):
    register_search(parser, 'author:Haspelmath provider:wals')


def run(args):
    try:
        fts.get_index(args.repos, must_exist=True)
    except ValueError:
        raise ParserError('Index does not exist. Run "glottolog searchindex" first!')
    count, results = fts.search(args.repos, args.query)
    with Table('ID', 'Author', 'Year', 'Title') as table:
        for res in results:
            table.append([res.id, res.author, res.year, res.title])
    print('({} matches)'.format(count))



# Python compatibility backports

import sys

__all__ = ['removesuffix']


if sys.version_info < (3, 9):
    def removesuffix(s: str, suffix: str) -> str:
        """See https://www.python.org/dev/peps/pep-0616/

        >>> removesuffix('spam.bib', '.bib')
        'spam'
        """
        return s[:-len(suffix)] if suffix and s.endswith(suffix) else s

else:  # pragma: no cover
    def removesuffix(s: str, suffix: str) -> str:
        """See https://www.python.org/dev/peps/pep-0616/

        >>> removesuffix('spam.bib', '.bib')
        'spam'
        """
        return s.removesuffix(suffix)



"""

"""
# Make sure we import all link providers:
from pyglottolog.links import *  # noqa: F401, F403
from pyglottolog.links.util import LinkProvider


def providers():
    return {cls.__name__.lower(): cls for cls in LinkProvider.__subclasses__()}


def register(parser):
    parser.add_argument('provider', nargs='*', help='|'.join(providers()))


def run(args):
    langs = list(args.repos.languoids())
    updated = set()
    for cls in LinkProvider.__subclasses__():
        name = cls.__name__.lower()
        if (not getattr(args, 'provider', None)) or (name in args.provider):
            args.log.info('updating {0} links ...'.format(name))
            i = 0
            for i, l in enumerate(cls(args.repos).iterupdated(langs), start=1):
                l.write_info()
                updated.add(l.id)
            args.log.info('... {0} done'.format(i))
    print('{0} languoids updated'.format(len(updated)))



import copy
import pathlib
import operator
import textwrap
import itertools
import functools

from termcolor import colored
from clldutils.iso_639_3 import ISO, download_tables


def sprint(text, *args, **kw):
    if not isinstance(text, str):
        text = '{0}'.format(text)
    if args:
        text = text.format(*args)
    color = kw.pop('color', None)
    attrs = kw.pop('attrs', None)
    if color or attrs:
        text = colored(text, color=color, attrs=attrs)
    print(text)


def wrap(text,
         line_as_paragraph=False,
         width=80,
         break_long_words=False,
         break_on_hyphens=False,
         **kw):
    kw.update(
        width=width, break_long_words=break_long_words, break_on_hyphens=break_on_hyphens)
    lines = []
    for line in text.split('\n'):
        if not line:
            lines.append('')
        else:
            lines.extend(textwrap.wrap(line, **kw))
            if line_as_paragraph:
                lines.append('')
    return '\n'.join(lines).strip()


def message(obj, msg):
    return '{0}: {1}'.format(colored('{0}'.format(obj), 'blue', attrs=['bold']), msg)


def get_iso(d):
    zips = sorted(
        list(pathlib.Path(d).glob('iso-639-3_Code_Tables_*.zip')),
        key=lambda p: p.name)
    if zips:
        return ISO(zips[-1])

    return ISO(download_tables(d))  # pragma: no cover


@functools.total_ordering
class Trigger(object):
    def __init__(self, field, type_, string):
        self.field = field
        self.type = type_
        self._string = string
        self.clauses = tuple(sorted([
            (False, w[4:].strip()) if w.startswith('NOT ') else (True, w.strip())
            for w in string.split(' AND ')]))

    def __eq__(self, other):
        # make triggers sortable so that we can easily group them by clauses.
        return self.clauses == other.clauses and self.cls == other.cls

    def __lt__(self, other):
        # make triggers sortable so that we can easily group them by clauses.
        return (self.clauses, self.cls) < (other.clauses, other.cls)

    @property
    def cls(self):
        return self.field, self.type

    def __call__(self, allkeys, keys_by_word):
        allkeys = set(allkeys)
        matching = copy.copy(allkeys)
        for isin, word in self.clauses:
            matching_for_clause = copy.copy(keys_by_word[word])
            if not isin:
                matching_for_clause = allkeys.difference(matching_for_clause)
            matching.intersection_update(matching_for_clause)
        return matching

    @staticmethod
    def format(label, triggers):
        trigs = [triggers] if isinstance(triggers, Trigger) else reversed(triggers)
        from_ = ';'.join(
            [' and '.join(
                [('' if c else 'not ') + w for c, w in t.clauses]) for t in trigs])
        return '%s (computerized assignment from "%s")' % (label, from_)

    @staticmethod
    def group(triggers):
        return [(clauses, list(trigs)) for clauses, trigs
                in itertools.groupby(sorted(triggers), lambda t: t.clauses)]


def group_first(iterable, groupkey=operator.itemgetter(0)):
    for key, group in itertools.groupby(iterable, groupkey):
        yield key, list(group)


def unique(iterable):
    seen = set()
    for item in iterable:
        if item not in seen:
            seen.add(item)
            yield item



# bibtex.py - bibtex file parsing/serialization

# TODO: make check fail on non-whitespace between entries (bibtex 'comments')

import typing
import pathlib
import functools
import collections
import unicodedata

from pybtex.database.input.bibtex import LowLevelParser, Parser, UndefinedMacro
from pybtex.scanner import PybtexSyntaxError
from pybtex.exceptions import PybtexError
from pybtex.textutils import whitespace_re
from pybtex.bibtex.utils import split_name_list
from pybtex.database import Person

from clldutils.path import memorymapped

FIELDORDER = ['author', 'editor', 'title', 'booktitle', 'journal',
              'school', 'publisher', 'address',
              'series', 'volume', 'number', 'pages', 'year', 'issn', 'url']


def load(filename, preserve_order=False, encoding=None):
    cls = collections.OrderedDict if preserve_order else dict
    return cls(iterentries(filename, encoding))


def identity(x):  # pragma: no cover
    return x


def iterentries_from_text(text, encoding='utf-8'):
    if hasattr(text, 'read'):
        text = text.read(-1)
    if not isinstance(text, str):
        text = text.decode(encoding)
    for entrytype, (bibkey, fields) in LowLevelParser(text):
        fields = {
            name.lower(): whitespace_re.sub(' ', ''.join(values)).strip()
            for name, values in fields}
        yield bibkey, (entrytype, fields)


def iterentries(filename: typing.Union[str, pathlib.Path], encoding=None)\
        -> typing.Generator[typing.Tuple[str, typing.Tuple[str, dict]], None, None]:
    encoding = encoding or 'utf8'
    with memorymapped(str(filename)) as source:
        try:
            for bibkey, (entrytype, fields) in iterentries_from_text(source, encoding):
                yield bibkey, (entrytype, fields)
        except PybtexSyntaxError as e:  # pragma: no cover
            debug_pybtex(source, e)


def debug_pybtex(source, e):  # pragma: no cover
    start, line, pos = e.error_context_info
    print('BIBTEX ERROR on line %d, last parsed lines:' % line)
    print(source[start:start + 500] + '...')
    raise e


def names(s):
    for name in split_name_list(s):
        try:
            yield Name.from_string(name)
        except PybtexError as e:  # pragma: no cover
            print(repr(e))


class Name(collections.namedtuple('Name', 'prelast last given lineage')):

    __slots__ = ()

    @classmethod
    def from_string(cls, name):
        person = Person(name)
        ntypes = ('prelast_names', 'last_names', 'first_names', 'middle_names', 'lineage_names')
        prelast, last, first, middle, lineage = (' '.join(getattr(person, part)) for part in ntypes)
        given = ' '.join(n for n in (first, middle) if n)
        return cls(prelast, last, given, lineage)


def save(entries, filename, sortkey, encoding='utf-8', normalize='NFC'):
    with open(str(filename), 'w', encoding=encoding, errors='strict') as fd:
        dump(entries, fd, sortkey, normalize)


def dump(entries, fd, sortkey=None, normalize='NFC'):
    assert sortkey in [None, 'bibkey']
    if sortkey is None:
        if isinstance(entries, collections.OrderedDict):  # pragma: no cover
            items = entries.items()
        elif isinstance(entries, dict):  # pragma: no cover
            raise ValueError('dump needs sortkey or ordered entries')
        else:
            items = entries
    else:  # elif sortkey == 'bibkey':
        items = (
            (bibkey, entries[bibkey])
            for bibkey in sorted(entries, key=lambda bibkey: bibkey.lower()))
    r"""Reserved characters (* -> en-/decoded by latexcodec)
    * #: \#
      $: \$
      %: \%
      ^: \^{} \textasciicircum
    * &: \&
    * _: \_
      {: \{
      }: \}
      ~: \~{} \textasciitilde
      \: \textbackslash{}
      <: \textless
      >: \textgreater
    """
    assert normalize in (None, '', 'NFC', 'NFKC', 'NFD', 'NFKD')
    if normalize:
        normalize = functools.partial(unicodedata.normalize, normalize)
    else:  # pragma: no cover
        normalize = identity
    fd.write(u'# -*- coding: utf-8 -*-\n')
    for bibkey, (entrytype, fields) in items:
        fd.write(u'@%s{%s' % (entrytype, bibkey))
        for k, v in fieldorder.itersorted(fields):
            fd.write(u',\n    %s = {%s}' % (k, normalize(v.strip())))
        fd.write(u'\n}\n' if fields else u',\n}\n')


class Ordering(dict):
    """Key order for iterating over dicts (unknown keys last alphabetic)."""

    _missing = float('inf')

    @classmethod
    def fromlist(cls, keys):
        """Define the order of keys as given."""
        return cls((k, i) for i, k in enumerate(keys))

    def itersorted(self, dct):
        """Iterate over dct (key, value) pairs in the defined order."""
        for key in sorted(dct, key=self._itersorted_key):
            yield key, dct[key]

    def _itersorted_key(self, key):
        return self[key], key

    def __missing__(self, key):
        return self._missing


fieldorder = Ordering.fromlist(FIELDORDER)


def check(filename, encoding=None):
    parser = CheckParser(encoding=encoding)
    parser.parse_file(str(filename))
    return parser.error_count


class CheckParser(Parser):
    """Unline LowLevelParser also parses names, macros, etc."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.error_count = 0

    def handle_error(self, error):  # pragma: no cover
        print('%r' % error)
        self.error_count += 1
        if not isinstance(error, UndefinedMacro):
            raise error

    def process_entry(self, *args, **kwargs):
        try:
            super().process_entry(*args, **kwargs)
        except PybtexError as e:  # pragma: no cover
            print(e)
            self.error_count += 1



"""
Update langsci.bib
"""
import re
import argparse
import collections

from clldutils.misc import slug
from clldutils.clilib import PathType
from pyglottolog.references.bibtex import iterentries


def register(parser):
    parser.add_argument(
        'bib',
        type=PathType(type='file'),
        help='path to updated langsci.bib')
    parser.add_argument(
        '--bibkey',
        default='langsci.bib',
        help=argparse.SUPPRESS)


def run(args):  # pragma: no cover
    args.repos.bibfiles[args.bibkey].update(args.bib, log=args.log)
    return
    #
    # FIXME: First update keys in old langsci.bib to the new ones based on
    #  - old citation key
    #  - title
    #
    titles = {}
    keys = {}
    by_stem = collections.defaultdict(dict)

    for key, (_, fields) in iterentries(args.bib):
        keys[key] = 1
        if re.search(r':[0-9]{2}:[0-9]$', key) and fields.get('title'):
            by_stem[':'.join(key.split(':')[:-1])][slug(fields['title'])] = key
        if fields.get('title'):
            titles[key] = slug(fields['title'])

    bib = args.repos.bibfiles[args.bibkey]

    def visitor(e):
        e.type = e.type.lower()
        lookup = e.key.lower()
        if lookup in keys:
            e.key = lookup
            return

        if lookup[-1] in 'abc':
            lookup = lookup[:-1]

        if lookup in by_stem:
            tslug = slug(e.fields.get('title', ''))
            if tslug in by_stem[lookup]:
                e.key = by_stem[lookup][tslug]
                return

        if ':ed:' in lookup:
            lookup = lookup.replace(':ed:', ':')
            if lookup in titles:
                if slug(e.fields.get('title', '')) == titles[lookup]:
                    e.key = lookup
                    return

        return True
    bib.visit(visitor)



import pathlib
import collections

import attr
from clldutils.misc import nfilter
from clldutils.inifile import INI

__all__ = [
    'AES', 'AESSource', 'Macroarea', 'DocumentType', 'LanguageType', 'LanguoidLevel',
    'Generic', 'Config']


class ConfigObject(object):
    """
    Factory to turn INI file sections into instances of `@attr.s` classes.
    """
    @classmethod
    def from_section(cls, cfg, section):
        try:
            fields = set(f.name for f in attr.fields(cls))
        except attr.exceptions.NotAnAttrsClassError:
            fields = None

        kw = {'name' if 'id' in cfg[section] else 'id': section}
        kw.update(cfg[section].items())
        return cls(**{k: v for k, v in kw.items() if fields is None or k in fields})


class Generic(ConfigObject):
    """
    Make config options available as attributes.
    """
    def __init__(self, **kw):
        for k, v in kw.items():
            if v in ['True', 'False']:
                v = eval(v)
            setattr(self, k, v)


@attr.s
class AES(ConfigObject):
    """
    AES status values

    .. seealso:: `<https://glottolog.org/langdoc/status>`_
    """
    # The attribute which is used for ordering objects of this type must come first:
    #: Sequential numeric value
    ordinal = attr.ib(converter=int)
    #: unique identifier (suitable as Python name, \
    #: see `<https://docs.python.org/3/reference/lexical_analysis.html#identifiers>`_)
    id = attr.ib()
    #: unique human-readable name
    name = attr.ib()
    #: corresponding status in the EGIDS scala
    egids = attr.ib()
    #: corresponding status in the UNESCO scala
    unesco = attr.ib()
    #: corresponding status in ElCat
    elcat = attr.ib()
    #: Glottolog reference ID linking to further information
    reference_id = attr.ib()
    icon = attr.ib(default=None)


@attr.s
class AESSource(ConfigObject):
    """
    Reference information for AES sources
    """
    id = attr.ib()  #:
    name = attr.ib()  #:
    url = attr.ib()  #:
    #: Glottolog reference ID linking to further information
    reference_id = attr.ib()
    pages = attr.ib(default=None)  #:


@attr.s
class Macroarea(ConfigObject):
    """
    Glottolog macroareas (see `<https://glottolog.org/meta/glossary#macroarea>`_)
    """
    id = attr.ib()  #:
    name = attr.ib()  #:
    description = attr.ib()  #:
    #: Glottolog reference ID linking to further information
    reference_id = attr.ib()


@attr.s
class DocumentType(ConfigObject):
    """
    Document types categorize Glottolog references
    """
    rank = attr.ib(converter=int)  #:
    id = attr.ib()  #:
    name = attr.ib()  #:
    description = attr.ib()  #:
    abbv = attr.ib()
    bibabbv = attr.ib()
    webabbr = attr.ib()
    triggers = attr.ib(converter=lambda s: nfilter(s.split('\n')))


@attr.s
class MEDType(ConfigObject):
    """
    MED (aka Descriptive Status) types (more coarse-grained document types)

    .. seealso:: `<https://glottolog.org/langdoc/status>`_
    """
    rank = attr.ib(converter=int)  #:
    id = attr.ib()  #:
    name = attr.ib()  #:
    description = attr.ib()  #:
    icon = attr.ib(default=None)


@attr.s
class LanguageType(ConfigObject):
    """
    Language types categorize languages.
    """
    id = attr.ib()  #:
    #: Glottocode of the pseudo-family that languages of this type are grouped in.
    pseudo_family_id = attr.ib()
    category = attr.ib()  #: category name for languages of this type
    description = attr.ib()  #:


@attr.s(hash=True)
class LanguoidLevel(ConfigObject):
    """
    Languoid levels describe the position of languoid nodes in the classification.

    :ivar name: alias for `id`
    """
    ordinal = attr.ib(converter=int)  #:
    id = attr.ib()  #:
    description = attr.ib()  #:

    @property
    def name(self):
        return self.id


def get_ini(fname, **kw):
    fname = pathlib.Path(fname)
    if not fname.exists():
        # For old-style (<=3.4) repository layout we ship the config data with pyglottolog:
        name = fname.name if fname.name != 'hhtype.ini' else 'document_types.ini'
        fname = pathlib.Path(__file__).parent / name
    assert fname.exists()
    return INI.from_file(fname, **kw)


class Config(collections.OrderedDict):
    """
    More convenient access to objects stored as sections in INI files

    This class makes objects (i.e. INI sections) accessible as values of a `dict`, keyed by an
    `id` attribute, which is infered from the `id` or `name` option of the section and, additonally,
    under as attribute named after `id`.
    """
    __defaults__ = {}

    @classmethod
    def from_ini(cls, fname, object_class):
        ini = get_ini(fname)
        d = collections.OrderedDict()
        for sec in ini.sections():
            obj = object_class.from_section(ini, sec)
            d[obj.id] = obj
        res = cls(**d)
        res.__defaults__ = ini['DEFAULT']
        return res

    def __getattribute__(self, item):
        if item in self:
            return self[item]
        return dict.__getattribute__(self, item)

    def get(self, item, default=None):
        if isinstance(item, str) and item in self:
            return self[item]
        if isinstance(item, ConfigObject) and getattr(item, 'id', None) in self:
            return self[item.id]
        for li in self.values():
            if any(getattr(li, attr, None) == item for attr in ['name', 'value', 'description']):
                return li
        if default:
            return default
        raise ValueError(item)



"""
Create an index pages listing and linking to all languoids of a specified level.
"""
import collections

from clldutils.misc import slug


def run(args):
    def make_index(level, languoids, repos):
        fname = dict(
            language='languages', family='families', dialect='dialects')[level.name]
        links = collections.defaultdict(dict)
        for lang in languoids:
            label = '{0.name} [{0.id}]'.format(lang)
            if lang.iso:
                label += '[%s]' % lang.iso
            links[slug(lang.name)[0]][label] = \
                lang.fname.relative_to(repos.languoids_path())

        with repos.languoids_path(fname + '.md').open('w', encoding='utf8') as fp:
            fp.write('## %s\n\n' % fname.capitalize())
            fp.write(' '.join(
                '[-%s-](%s_%s.md)' % (i.upper(), fname, i) for i in sorted(links.keys())))
            fp.write('\n')

        for i, langs in links.items():
            with repos.languoids_path(
                    '%s_%s.md' % (fname, i)).open('w', encoding='utf8') as fp:
                for label in sorted(langs.keys()):
                    fp.write('- [%s](%s)\n' % (label, langs[label]))

    langs = list(args.repos.languoids())
    for level in args.repos.languoid_levels.values():
        make_index(level, [lang for lang in langs if lang.level == level], args.repos)



"""
Programmatic access to Glottolog data.
"""
import re
import typing
import pathlib
import contextlib
import collections

import pycldf.util
from csvw import TableGroup, Column
from clldutils.path import walk, git_describe
from clldutils.misc import lazyproperty
from clldutils.apilib import API
from clldutils.jsonlib import load
import clldutils.iso_639_3
import pycountry
from termcolor import colored
from tqdm import tqdm

from . import util
from . import languoids as lls
from . import references
from . import config
from .languoids import models

__all__ = ['Glottolog']

ISO_CODE_PATTERN = re.compile('[a-z]{3}$')


class Cache(dict):
    def __init__(self):
        super().__init__()
        self._lineage = {}

    def __bool__(self):
        return True

    def add(self, directory: pathlib.Path, api) -> lls.Languoid:
        if directory.name not in self:
            lang = lls.Languoid.from_dir(directory, nodes=self._lineage, _api=api)
            self._lineage[lang.id] = (lang.name, lang.id, lang.level)
            self[lang.id] = lang
            if lang.iso:
                self[lang.iso] = lang
        else:
            lang = self[directory.name]
        return lang


class Glottolog(API):
    """
    API to access Glottolog data

    This class provides (read and write) access to a local copy of the Glottolog data, which can
    be obtained as explained in the `README <https://github.com/glottolog/pyglottolog#install>`_
    """
    countries = [models.Country(c.alpha_2, c.name) for c in pycountry.countries]

    def __init__(self, repos='.', *, cache: bool = False):
        """
        :param repos: Path to a copy of `<https://github.com/glottolog/glottolog>`_
        :param cache: Indicate whether to cache `Languoid` objects or not. If `True`, the API must \
        be used read-only.
        """
        API.__init__(self, repos=repos)
        #: Absolute path to the copy of the data repository:
        self.repos: pathlib.Path = pathlib.Path.cwd() / self.repos
        #: Absolute path to the `tree` directory in the repos.
        self.tree: pathlib.Path = self.repos / 'languoids' / 'tree'
        if not self.tree.exists():
            raise ValueError('repos dir %s missing tree dir: %s' % (self.repos, self.tree))
        if not self.repos.joinpath('references').exists():
            raise ValueError('repos dir %s missing references subdir' % (self.repos,))
        self.cache = Cache() if cache else None

    def __str__(self):
        return '<Glottolog repos {0} at {1}>'.format(git_describe(self.repos), self.repos)

    def describe(self) -> str:
        return git_describe(self.repos)

    def references_path(self, *comps: str):
        """
        Path within the `references` directory of the repos.
        """
        return self.repos.joinpath('references', *comps)

    def languoids_path(self, *comps):
        """
        Path within the `languoids` directory of the repos.
        """
        return self.repos.joinpath('languoids', *comps)

    def build_path(self, *comps: str) -> pathlib.Path:
        build_dir = self.repos.joinpath('build')
        if not build_dir.exists():
            build_dir.mkdir()  # pragma: no cover
        return build_dir.joinpath(*comps)

    @contextlib.contextmanager
    def cache_dir(self, name):
        d = self.build_path(name)
        if not d.exists():
            d.mkdir()
        yield d

    def _cfg(self, name, cls=None):
        return config.Config.from_ini(
            self.path('config', name + '.ini'), object_class=cls or config.Generic)

    @lazyproperty
    def aes_status(self) -> typing.Dict[str, config.AES]:
        """
        :rtype: mapping with :class:`config.AES` values.
        """
        return self._cfg('aes_status', cls=config.AES)

    @lazyproperty
    def aes_sources(self) -> typing.Dict[str, config.AESSource]:
        """
        :rtype: mapping with :class:`config.AESSource` values
        """
        return self._cfg('aes_sources', cls=config.AESSource)

    @lazyproperty
    def document_types(self) -> typing.Dict[str, config.DocumentType]:
        """
        :rtype: mapping with :class:`config.DocumentType` values
        """
        return self._cfg('document_types', cls=config.DocumentType)

    @lazyproperty
    def med_types(self) -> typing.Dict[str, config.MEDType]:
        """
        :rtype: mapping with :class:`config.MEDType` values
        """
        return self._cfg('med_types', cls=config.MEDType)

    @lazyproperty
    def macroareas(self) -> typing.Dict[str, config.Macroarea]:
        """
        :rtype: mapping with :class:`config.Macroarea` values
        """
        return self._cfg('macroareas', cls=config.Macroarea)

    @lazyproperty
    def language_types(self) -> typing.Dict[str, config.LanguageType]:
        """
        :rtype: mapping with :class:`config.LanguageType` values
        """
        return self._cfg('language_types', cls=config.LanguageType)

    @lazyproperty
    def languoid_levels(self) -> typing.Dict[str, config.LanguoidLevel]:
        """
        :rtype: mapping with :class:`config.LanguoidLevel` values
        """
        return self._cfg('languoid_levels', cls=config.LanguoidLevel)

    @lazyproperty
    def editors(self) -> typing.Dict[str, config.Generic]:
        """
        Metadata about editors of Glottolog

        :rtype: mapping with :class:`config.Generic` values
        """
        return self._cfg('editors')

    @lazyproperty
    def publication(self) -> typing.Dict[str, config.Generic]:
        """
        Metadata about the Glottolog publication

        :rtype: mapping with :class:`config.Generic` values
        """
        return self._cfg('publication')

    @lazyproperty
    def iso(self) -> clldutils.iso_639_3.ISO:
        """
        :return: `clldutils.iso_639_3.ISO` instance, fed with the data of the latest \
        ISO code table zip found in the `build` directory.
        """
        return util.get_iso(self.build_path())

    @property
    def ftsindex(self) -> pathlib.Path:
        """
        Directory within `build` where the FullTextSearch index is created.
        """
        return self.build_path('whoosh')

    @lazyproperty
    def _tree_dirs(self):
        return list(walk(self.tree, mode='dirs'))

    @property
    def glottocodes(self) -> models.Glottocodes:
        """
        Registry of Glottocodes.
        """
        return models.Glottocodes(self.languoids_path('glottocodes.json'))

    def languoid(self, id_: typing.Union[str, lls.Languoid]) -> lls.Languoid:
        """
        Retrieve a languoid specified by language code.

        :param id_: Glottocode or ISO code.
        """
        if isinstance(id_, lls.Languoid):
            return id_

        if self.cache and id_ in self.cache:
            return self.cache[id_]

        if ISO_CODE_PATTERN.match(id_):
            for d in self._tree_dirs if self.cache else walk(self.tree, mode='dirs'):
                if self.cache:
                    l_ = self.cache.add(d, self)
                else:
                    l_ = lls.Languoid.from_dir(d, _api=self)
                if l_.iso_code == id_:
                    return l_
        else:
            for d in self._tree_dirs if self.cache else walk(self.tree, mode='dirs'):
                l_ = None
                if self.cache:
                    # If we cache Languoids, we might as well instantiate the ones we traverse:
                    l_ = self.cache.add(d, self)
                if d.name == id_:
                    if self.cache:
                        return l_
                    return lls.Languoid.from_dir(d, _api=self)

    def languoids(
            self,
            ids: set = None,
            maxlevel: typing.Union[int, config.LanguoidLevel, str] = None,
            exclude_pseudo_families: bool = False
    ) -> typing.Generator[lls.Languoid, None, None]:
        """
        Yields languoid objects.

        :param ids: `set` of Glottocodes to limit the result to. This is useful to increase \
        performance, since INI file reading can be skipped for languoids not listed.
        :param maxlevel: Numeric maximal nesting depth of languoids, or Languoid.level.
        :param exclude_pseudo_families: Flag signaling whether to exclude pseud families, \
        i.e. languoids from non-genealogical trees.
        """
        is_max_level_int = isinstance(maxlevel, int)
        # Non-numeric levels are interpreted as `Languoid.level` descriptors.
        if not is_max_level_int:
            maxlevel = self.languoid_levels.get(maxlevel or 'dialect')

        # Since we traverse the tree topdown, we can cache a mapping of Languoid.id to triples
        # (name, id, level) for populating `Languoid.lineage`.
        nodes = {}
        for d in self._tree_dirs if self.cache else walk(self.tree, mode='dirs'):
            if ids is None or d.name in ids:
                if self.cache:
                    lang = self.cache.add(d, self)
                else:
                    lang = lls.Languoid.from_dir(d, nodes=nodes, _api=self)
                if (is_max_level_int and len(lang.lineage) <= maxlevel) \
                        or ((not is_max_level_int) and lang.level <= maxlevel):
                    if (not exclude_pseudo_families) or not lang.category.startswith('Pseudo'):
                        yield lang

    def languoids_by_code(self, nodes=None) -> typing.Dict[str, lls.Languoid]:
        """
        Returns a `dict` mapping the three major language code schemes
        (Glottocode, ISO code, and Harald's NOCODE_s) to Languoid objects.
        """
        res = {}
        for lang in (self.languoids() if nodes is None else nodes.values()):
            res[lang.id] = lang
            if lang.hid:
                res[lang.hid] = lang
            if lang.iso:
                res[lang.iso] = lang
        return res

    def ascii_tree(self, start: typing.Union[str, lls.Languoid], maxlevel=None):
        """
        Prints an ASCII representation of the languoid tree starting at `start` to `stdout`.
        """
        _ascii_node(
            self.languoid(start),
            0,
            True,
            self.languoid_levels.get(maxlevel, maxlevel) if maxlevel else None,
            '',
            self.languoid_levels)

    def newick_tree(
            self,
            start: typing.Union[None, str, lls.Languoid] = None,
            template: str = None,
            nodes=None,
            maxlevel: typing.Union[int, config.LanguoidLevel] = None
    ) -> str:
        """
        Returns the Newick representation of a (set of) Glottolog classification tree(s).

        :param start: Root languoid of the tree (or `None` to return the complete classification).
        :param template: Python format string accepting the `Languoid` instance as single \
        variable named `l`, used to format node labels.
        """
        template = template or lls.Languoid._newick_default_template
        if start:
            return self.languoid(start).newick_node(
                template=template, nodes=nodes, maxlevel=maxlevel, level=1).newick + ';'
        if nodes is None:
            nodes = collections.OrderedDict((lang.id, lang) for lang in self.languoids())
        trees = []
        for lang in nodes.values():
            if not lang.lineage and not lang.category.startswith('Pseudo '):
                ns = lang.newick_node(
                    nodes=nodes, template=template, maxlevel=maxlevel, level=1).newick
                if lang.level == self.languoid_levels.language:
                    # An isolate: we wrap it in a pseudo-family with the same name and ID.
                    fam = lls.Languoid.from_name_id_level(
                        lang.dir.parent, lang.name, lang.id, 'family', _api=self)
                    ns = '({0}){1}:1'.format(ns, template.format(l=fam))  # noqa: E741
                trees.append('{0};'.format(ns))
        return '\n'.join(trees)

    @lazyproperty
    def bibfiles(self) -> references.BibFiles:
        """
        Access reference data by BibFile.

        :rtype: :class:`references.BibFiles`
        """
        return references.BibFiles.from_path(self.references_path(), api=self)

    def refs_by_languoid(self, *bibfiles, **kw):
        nodes = kw.get('nodes')
        if bibfiles:
            bibfiles = [
                bib if isinstance(bib, references.BibFile) else self.bibfiles[bib]
                for bib in bibfiles]
        else:
            bibfiles = self.bibfiles
        all_ = {}
        languoids_by_code = self.languoids_by_code(
            nodes or {lang.id: lang for lang in self.languoids()})
        res = collections.defaultdict(list)
        for bib in tqdm(bibfiles):
            for entry in bib.iterentries():
                all_[entry.id] = entry
                for lang in entry.languoids(languoids_by_code)[0]:
                    res[lang.id].append(entry)
        return res, all_

    @lazyproperty
    def hhtypes(self):
        # Note: The file `hhtype.ini` does not exist anymore. This is fixed in HHTypes, when
        # calling `config.get_ini`. Only used when compiling monster.bib.
        return references.HHTypes(self.references_path('hhtype.ini'))

    @lazyproperty
    def triggers(self):
        res = {'inlg': [], 'lgcode': []}
        for lang in self.languoids():
            for type_ in res:
                if lang.cfg.has_option('triggers', type_):
                    label = '%s [%s]' % (lang.name, lang.hid or lang.id)
                    res[type_].extend([util.Trigger(type_, label, text)
                                       for text in lang.cfg.getlist('triggers', type_)])
        return res

    @lazyproperty
    def macroarea_map(self):
        res = {}
        for lang in self.languoids():
            ma = lang.macroareas[0].name if lang.macroareas else ''
            res[lang.id] = ma
            if lang.iso:
                res[lang.iso] = ma
            if lang.hid:
                res[lang.hid] = ma
        return res

    @property
    def current_editors(self):
        return sorted([e for e in self.editors.values() if e.current], key=lambda e: int(e.ord))

    def write_languoids_table(self, outdir, version=None):
        version = version or self.describe()
        out = outdir / 'glottolog-languoids-{0}.csv'.format(version)
        md = outdir / (out.name + '-metadata.json')
        tg = TableGroup.fromvalue({
            "@context": "http://www.w3.org/ns/csvw",
            "dc:version": version,
            "dc:bibliographicCitation":
                "{0}. "
                "{1} [Data set]. "
                "Zenodo. https://doi.org/{2}".format(
                    ' & '.join([e.name for e in self.current_editors]),
                    self.publication.zenodo.title_format.format('(Version {0})'.format(version)),
                    self.publication.zenodo.doi,
                ),
            "tables": [load(pycldf.util.pkg_path('components', 'LanguageTable-metadata.json'))],
        })
        tg.tables[0].url = out.name
        for col in [
            dict(name='LL_Code'),
            dict(name='Classification', separator='/'),
            dict(name='Family_Glottocode'),
            dict(name='Family_Name'),
            dict(name='Language_Glottocode'),
            dict(name='Language_Name'),
            dict(name='Level', datatype=dict(base='string', format='family|language|dialect')),
            dict(name='Status'),
        ]:
            tg.tables[0].tableSchema.columns.append(Column.fromvalue(col))

        langs = []
        for lang in self.languoids():
            lid, lname = None, None
            if lang.level == self.languoid_levels.language:
                lid, lname = lang.id, lang.name
            elif lang.level == self.languoid_levels.dialect:
                for lname, lid, level in reversed(lang.lineage):
                    if level == self.languoid_levels.language:
                        break
                else:  # pragma: no cover
                    raise ValueError
            langs.append(dict(
                ID=lang.id,
                Name=lang.name,
                Macroarea=lang.macroareas[0].name if lang.macroareas else None,
                Latitude=lang.latitude,
                Longitude=lang.longitude,
                Glottocode=lang.id,
                ISO639P3code=lang.iso,
                LL_Code=lang.identifier.get('multitree'),
                Classification=[c[1] for c in lang.lineage],
                Language_Glottocode=lid,
                Language_Name=lname,
                Family_Name=lang.lineage[0][0] if lang.lineage else None,
                Family_Glottocode=lang.lineage[0][1] if lang.lineage else None,
                Level=lang.level.name,
                Status=lang.endangerment.status.name if lang.endangerment else None,
            ))

        tg.to_file(md)
        tg.tables[0].write(langs, fname=out)
        return md, out


def _ascii_node(n, level, last, maxlevel, prefix, levels):
    nlevel = levels.get(n.level)
    if maxlevel:
        if (isinstance(maxlevel, config.LanguoidLevel) and nlevel > maxlevel) or \
                (not isinstance(maxlevel, config.LanguoidLevel) and level > maxlevel):
            return
    s = '\u2514' if last else '\u251c'
    s += '\u2500 '

    if not level:
        for i, node in enumerate(n.ancestors):
            util.sprint('{0}{1}{2} [{3}]', prefix, s if i else '', node.name, node.id)
            prefix = '   ' + prefix

    nprefix = prefix + ('   ' if last else '\u2502  ')

    color = 'red' if not level else (
        'green' if nlevel == levels.language else (
            'blue' if nlevel == levels.dialect else None))

    util.sprint(
        '{0}{1}{2} [{3}]',
        prefix,
        s if level else (s if n.ancestors else ''),
        colored(n.name, color) if color else n.name,
        colored(n.id, color) if color else n.id)
    for i, c in enumerate(sorted(n.children, key=lambda nn: nn.name)):
        _ascii_node(c, level + 1, i == len(n.children) - 1, maxlevel, nprefix, levels)



"""
Print the classification tree starting at a specific languoid.
"""
from clldutils.clilib import ParserError

from pyglottolog.languoids import Languoid
from pyglottolog.util import sprint
from pyglottolog.cli_util import get_languoid


def register(parser):
    parser.add_argument(
        'root',
        metavar='<GLOTTOCODE>|<ISO-CODE>',
        nargs='?',
        help="Root node for the tree",
        default=None,
    )
    parser.add_argument(
        '--newick',
        help='Serialize tree in Newick format',
        default=False,
        action='store_true',
    )
    parser.add_argument(
        '--maxlevel',
        help='Maximal branch length, or maximal level of languoids to include',
        default=None,
    )
    parser.add_argument(
        '--template',
        help="""node label template, used to control the labels in the Newick string.
Values for this option must be valid python format strings expecting a single
template variable `l` which points to the Languoid instance.
In addition to Languoid attributes and properties specified as "{{l.<attr>}}",
e.g. "{{l.id}}" for the Glottocode of a Languoid, the following custom format specs
can be used:
{}""".format(
            '\n'.join('"l:{0}": {1[1]}; '.format(k, v) for k, v in Languoid._format_specs.items())),
        default=None,
    )


def run(args):
    root = get_languoid(args, args.root)

    if args.maxlevel is not None:
        try:
            args.maxlevel = int(args.maxlevel)
        except Exception:
            args.maxlevel = args.repos.languoid_levels[args.maxlevel] \
                if args.maxlevel in args.repos.languoid_levels else None
    if args.newick:
        sprint(args.repos.newick_tree(root, template=args.template, maxlevel=args.maxlevel))
    else:
        if not root:
            raise ParserError('Root languoid must be specified for non-newick serialization!')
        args.repos.ascii_tree(root, maxlevel=args.maxlevel)



"""
Add missing macroareas for dialects.

By default, dialects are assumed to share the macroarea of their language.
"""


def run(args):
    langs = {lang.id: lang for lang in args.repos.languoids()}
    updated = []

    for dialect in langs.values():
        if dialect.level == args.repos.languoid_levels['dialect'] and not dialect.macroareas:
            for _, lid, _ in reversed(dialect.lineage):
                language = langs[lid]
                if language.level == args.repos.languoid_levels['language']:
                    break
            else:  # pragma: no cover
                raise ValueError()
            dialect.macroareas = [ma for ma in language.macroareas]
            updated.append(dialect)

    for lang in updated:
        lang.write_info()

    print('{} dialects updated'.format(len(updated)))



import os
import re
import typing
import pathlib
import datetime
import warnings
import functools
import configparser

from clldutils.inifile import INI
from newick import Node

from .models import (
    Glottocode, Country, Reference, Endangerment, Link,
    ClassificationComment, EthnologueComment, ISORetirement,
)
from pyglottolog import config

__all__ = ['Languoid']

INFO_FILENAME = 'md.ini'

ISO_8601_INTERVAL = re.compile(
    r'(?P<start_sign>[+-]?)'
    r'(?P<start_date>\d{1,4}-\d{2}-\d{2})'
    r'/'
    r'(?P<end_sign>[+-]?)'
    r'(?P<end_date>\d{1,4}-\d{2}-\d{2})',
    flags=re.ASCII)


@functools.total_ordering
class Languoid(object):
    """
    Info on languoids is encoded in the INI files and in the directory hierarchy of
    :attr:`pyglottolog.Glottolog.tree`.
    This class provides access to all of it.

    **Languoid formatting**:

    :ivar _format_specs: A `dict` mapping custom format specifiers to conversion functions. Usage:

    .. code-block:: python

        >>> l = Languoid.from_name_id_level(pathlib.Path('.'), 'N(a,m)e', 'abcd1234', 'language')
        >>> '{0:newick_name}'.format(l)
        'N{a/m}e'

    .. seealso::

        `<https://www.python.org/dev/peps/pep-3101/#format-specifiers>`_ and
        `<https://www.python.org/dev/peps/pep-3101/#controlling-formatting-on-a-per-type-basis>`_
    """
    section_core = 'core'

    def __init__(
            self,
            cfg: INI,
            lineage: typing.Union[None, typing.List[typing.Tuple[str, str, str]]] = None,
            id_: typing.Union[None, str] = None,
            directory: typing.Union[None, pathlib.Path] = None,
            tree: typing.Union[None, pathlib.Path] = None,
            _api=None):
        """
        Refer to the factory methods for typical use cases of instantiating a `Languoid`:

        - :meth:`Languoid.from_dir`
        - :meth:`Languoid.from_id_name_level`

        :param cfg: `INI` instance storing the languoid's metadata.
        :param lineage: list of ancestors (from root to this languoid).
        :param id_: Glottocode for the languoid (or `None`, if `directory` is passed).
        :param _api: Some properties require access to config data which is accessed through a \
        `Glottolog` API instance.
        """
        assert (id_ and tree) or directory
        if id_ is None:
            id_ = Glottocode(directory.name)
        lineage = lineage or []
        assert all(Glottocode.pattern.match(id) for _, id, _ in lineage)
        self.lineage = [
            (name, id, _api.languoid_levels.get(level) if _api else level)
            for name, id, level in lineage]
        self.cfg = cfg
        self.dir = directory or tree.joinpath(*[id for name, id, _ in self.lineage])
        self._id = id_
        self._api = _api

    @classmethod
    def from_dir(cls, directory: pathlib.Path, nodes=None, _api=None, **kw):
        """
        Create a `Languoid` from a directory, named with the Glottocode and containing `md.ini`.

        This method is used by :class:`pyglottolog.Glottolog` to read `Languoid`s from the
        repository's `languoids/tree` directory.
        """
        if _api and _api.cache and directory.name in _api.cache:
            return _api.cache[directory.name]

        if nodes is None:
            nodes = {}
        cfg = INI.from_file(directory.joinpath(INFO_FILENAME), interpolation=None)

        lineage = []
        for parent in directory.parents:
            id_ = parent.name
            assert id_ != directory.name
            if not Glottocode.pattern.match(id_):
                # we ignore leading non-languoid-dir path components.
                break

            if id_ not in nodes:
                l_ = Languoid.from_dir(parent, nodes=nodes, _api=_api, **kw)
                nodes[id_] = (l_.name, l_.id, l_.level)
            lineage.append(nodes[id_])

        res = cls(cfg, list(reversed(lineage)), directory=directory, _api=_api, **kw)
        nodes[res.id] = (res.name, res.id, res.level)
        return res

    @classmethod
    def from_name_id_level(cls, tree, name, id, level, **kw):
        """
        This method is used in `pyglottolog.lff` to instantiate `Languoid` s for new nodes
        encountered in "lff"-format trees.
        """
        cfg = INI(interpolation=None)
        cfg.read_dict(dict(core=dict(name=name)))
        res = cls(cfg, kw.pop('lineage', []), id_=Glottocode(id), tree=tree)
        for k, v in kw.items():
            setattr(res, k, v)
        # Note: Setting the level behaves differently when `_api` is available, so must be done
        # after all other attributes are initialized.
        res.level = level
        return res

    # We provide a couple of node label format specifications which can be used when serializing
    # trees in newick format.
    _format_specs = {
        'newick_name': (
            lambda l: l.name.replace(
                ',', '/').replace('(', '{').replace(')', '}').replace("'", "''"),
            "Languoid name with special newick characters replaced"),
        'newick_level': (
            lambda l: '-l-' if getattr(l.level, 'id', l.level) == 'language' else '',
            "Languoid level in case of languages"),
        'newick_iso': (
            lambda l: '[{0}]'.format(l.iso) if l.iso else '',
            "Bracketed ISO code or nothing"),
    }
    _newick_default_template = "'{l:newick_name} [{l.id}]{l:newick_iso}{l:newick_level}'"

    def __format__(self, format_spec):
        if format_spec in self._format_specs:
            return self._format_specs[format_spec][0](self)
        return object.__format__(self, format_spec)

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return self.id == other.id

    def __lt__(self, other):
        """
        To allow Languoid lists to be sorted, we implement a simple ordering by Glottocode.
        """
        return self.id < other.id

    def __repr__(self):
        return '<%s %s>' % (getattr(self.level, 'name', self.level).capitalize(), self.id)

    def __str__(self):
        return '%s [%s]' % (self.name, self.id)

    def _set(self, key, value, section=None):
        section = section or self.section_core
        if value is None and key in self.cfg[section]:
            del self.cfg[section][key]
        else:
            self.cfg.set(section, key, value)

    def _get(self, key, type_=None):
        res = self.cfg.get(self.section_core, key, fallback=None)
        if type_ and res:
            return type_(res)
        return res

    def newick_node(self, nodes=None, template=None, maxlevel=None, level=0) -> Node:
        """
        Return a `newick.Node` representing the subtree of the Glottolog classification starting
        at the languoid.

        :param template: Python format string accepting the `Languoid` instance as single \
        variable named `l`, used to format node labels.
        """
        template = template or self._newick_default_template
        n = Node(name=template.format(l=self), length='1')  # noqa: E741

        children = self.children if nodes is None else self.children_from_nodemap(nodes)
        for nn in sorted(children, key=lambda nn: nn.name):
            if maxlevel:
                if (isinstance(maxlevel, config.LanguoidLevel) and nn.level > maxlevel) or \
                        (not isinstance(maxlevel, config.LanguoidLevel) and level > maxlevel):
                    continue
            n.add_descendant(
                nn.newick_node(nodes=nodes, template=template, maxlevel=maxlevel, level=level + 1))
        return n

    def write_info(self, outdir: typing.Union[None, pathlib.Path] = None):
        """
        Write `Languoid` metadata as INI file to `outdir/<INFO_FILENAME>`.
        """
        outdir = outdir or self.dir
        if not isinstance(outdir, pathlib.Path):
            outdir = pathlib.Path(outdir)
        if outdir.name != self.id:
            outdir = outdir.joinpath(self.id)
        if not outdir.exists():
            outdir.mkdir()
        fname = outdir.joinpath(INFO_FILENAME)
        self.cfg.write(fname)
        if os.linesep == '\n':
            with fname.open(encoding='utf8') as fp:
                text = fp.read()
            with fname.open('w', encoding='utf8') as fp:
                fp.write(text.replace('\n', '\r\n'))
        return fname

    # -------------------------------------------------------------------------
    # Accessing info of a languoid
    # -------------------------------------------------------------------------
    @property
    def glottocode(self):
        """Alias for `id`"""
        return self._id

    @property
    def id(self):
        return self._id

    @property
    def category(self):
        """
        Languoid category.

        - Category name from :class:`pyglottolog.config.LanguageType` for languoids of level \
          "language",
        - `"Family"` or `"Pseudo Family"` for families,
        - `"Dialect"` for dialects.
        """
        # Computing the category requires access to config data:
        if self._api:
            pseudo_families = {
                c.pseudo_family_id: c.category for c in self._api.language_types.values()}
            fid = self.lineage[0][1] if self.lineage else None
            if self.level == self._api.languoid_levels.language:
                return pseudo_families.get(fid, self._api.language_types['spoken_l1'].category)
            cat = self.level.name.capitalize()
            if self.level == self._api.languoid_levels.family:
                if self.id.startswith('unun9') or \
                        self.id in pseudo_families or fid in pseudo_families:
                    cat = 'Pseudo ' + cat
            return cat

    @property
    def isolate(self) -> bool:
        """
        Flag signaling whether the languoid is an isolate, i.e. has level "language" and is not
        member of a family.
        """
        return getattr(self.level, 'id', self.level) == 'language' and not self.lineage

    def children_from_nodemap(self, nodes):
        # A faster alternative to `children` when the relevant languoids have already been
        # read from disc.
        return [nodes[d.name] for d in self.dir.iterdir() if d.is_dir()]

    def descendants_from_nodemap(self, nodes, level=None):
        if isinstance(level, str):
            level = self._api.languoid_levels.get(level)
        return [
            n for n in nodes.values() if
            n.lineage and self.id in [li[1] for li in n.lineage] and  # noqa: W504
            ((level is None) or n.level == level)]

    @property
    def children(self) -> typing.List['Languoid']:
        """
        List of direct descendants of the languoid in the classification tree.

        .. note::

            Using this on many languoids can be slow, because the directory tree may be traversed
            and INI files read multiple times. To circumvent this problem, you may use a read-only
            :class:`pyglottolog.Glottolog` instance, by passing `cache=True` at initialization.
        """
        return [Languoid.from_dir(d, _api=self._api) for d in self.dir.iterdir() if d.is_dir()]

    def ancestors_from_nodemap(self, nodes):
        # A faster alternative to `ancestors` when the relevant languoids have already
        # been read from disc.
        return [nodes[lineage[1]] for lineage in self.lineage]

    def iter_ancestors(self):
        for parent in self.dir.parents:
            id_ = parent.name
            if Glottocode.pattern.match(id_):
                yield Languoid.from_dir(parent, _api=self._api)
            else:
                # we ignore leading non-languoid-dir path components.
                break

    def iter_descendants(self):
        for child in self.children:
            yield child
            yield from child.iter_descendants()

    @property
    def ancestors(self) -> typing.List['Languoid']:
        """
        List of ancestors of the languoid in the classification tree, from root (i.e. top-level
        family) to parent node.

        .. note::

            Using this on many languoids can be slow, because the directory tree may be traversed
            and INI files read multiple times. To circumvent this problem, you may use a read-only
            :class:`pyglottolog.Glottolog` instance, by passing `cache=True` at initialization.
        """
        return list(reversed(list(self.iter_ancestors())))

    @property
    def parent(self) -> typing.Union['Languoid', None]:
        """
        Parent languoid or `None`.

        .. note::

            Using this on many languoids can be slow, because the directory tree may be traversed
            and INI files read multiple times. To circumvent this problem, you may use a read-only
            :class:`pyglottolog.Glottolog` instance, by passing `cache=True` at initialization.
        """
        try:
            return next(self.iter_ancestors())
        except StopIteration:
            return

    @property
    def family(self) -> typing.Union['Languoid', None]:
        """
        Top-level family the languoid belongs to or `None`.

        .. note::

            Using this on many languoids can be slow, because the directory tree may be traversed
            and INI files read multiple times. To circumvent this problem, you may use a read-only
            :class:`pyglottolog.Glottolog` instance, by passing `cache=True` at initialization.
        """
        return self.ancestors[0] if self.lineage else None

    @property
    def names(self) -> typing.Dict[str, list]:
        """
        A `dict` mapping alternative name providers to `list` s of alternative names for the
        languoid by the given provider.
        """
        if 'altnames' in self.cfg:
            return {k: self.cfg.getlist('altnames', k) for k in self.cfg['altnames']}
        return {}

    def add_name(self, name, type_='glottolog'):
        names = self.cfg.getlist('altnames', type_)
        if name not in names:
            self.cfg.set('altnames', type_, sorted(names + [name]))

    def update_names(self, names, type_='glottolog'):
        new = set(names)
        if new != set(self.cfg.getlist('altnames', type_)):
            self.cfg.set('altnames', type_, sorted(new))
            return True
        return False

    @property
    def identifier(self) -> typing.Union[dict, configparser.SectionProxy]:
        if 'identifier' in self.cfg:
            return self.cfg['identifier']
        return {}

    @property
    def sources(self) -> typing.List[Reference]:
        """
        List of Glottolog references linked to the languoid

        :rtype: :class:`pyglottolog.references.Reference`
        """
        if self.cfg.has_option('sources', 'glottolog'):
            return Reference.from_list(self.cfg.getlist('sources', 'glottolog'))
        return []

    @sources.setter
    def sources(self, refs):
        assert all(isinstance(r, Reference) for r in refs)
        self.cfg.set('sources', 'glottolog', ['{0}'.format(ref) for ref in refs])

    @property
    def endangerment(self) -> typing.Union[None, Endangerment]:
        """
        Endangerment information about the languoid.

        :rtype: :class:`Endangerment`
        """
        if ('endangerment' in self.cfg) and self._api:
            kw = {k: v for k, v in self.cfg['endangerment'].items()}
            kw['status'] = self._api.aes_status.get(kw['status'])
            if kw['source'] in self._api.aes_sources:
                kw['source'] = self._api.aes_sources[kw['source']]
            else:
                ref = Reference.from_string(kw['source'])
                kw['source'] = config.AESSource(
                    id=ref.key,
                    name=None,
                    url=None,
                    reference_id=ref.key,
                    pages=ref.pages)
            return Endangerment(**kw)

    @property
    def classification_comment(self) -> typing.Union[None, ClassificationComment]:
        """
        Classification information about the languoid.

        :rtype: :class:`ClassificationComment`
        """
        if 'classification' in self.cfg:
            cfg = self.cfg['classification']
            return ClassificationComment(
                family=cfg.get('family'),
                familyrefs=self.cfg.getlist('classification', 'familyrefs'),
                sub=cfg.get('sub'),
                subrefs=self.cfg.getlist('classification', 'subrefs'))

    @property
    def ethnologue_comment(self) -> typing.Union[None, EthnologueComment]:
        """
        Commentary about the classification of the languoid in Ethnologue.

        :rtype: :class:`EthnologueComment`
        """
        section = 'hh_ethnologue_comment'
        if section in self.cfg:
            return EthnologueComment(**self.cfg[section])

    @property
    def macroareas(self) -> typing.List[config.Macroarea]:
        """
        :rtype: `list` of :class:`config.Macroarea`
        """
        if self._api:
            return [
                self._api.macroareas.get(n)
                for n in self.cfg.getlist(self.section_core, 'macroareas')]
        return []

    @macroareas.setter
    def macroareas(self, value):
        if self._api:
            assert isinstance(value, (list, tuple)) \
                and all(self._api.macroareas.get(n) for n in value)
            self._set('macroareas', [ma.name for ma in value])

    @property
    def timespan(self, _date_format='%Y-%m-%d'):
        value = self.cfg.get(self.section_core, 'timespan',
                             fallback=None)
        if not value:
            return None
        value = value.strip()
        ma = ISO_8601_INTERVAL.fullmatch(value)
        if ma is None:
            raise ValueError('invalid interval', value)  # pragma: no cover

        dates = ma.group('start_date', 'end_date')

        def fix_date(d, year_tmpl='{:04d}'):
            year, sep, rest = d.partition('-')
            assert year and sep and rest
            year = year_tmpl.format(int(year))
            return '{}{}{}'.format(year, sep, rest)

        dates = map(fix_date, dates)
        dates = [datetime.datetime.strptime(d, _date_format).date() for d in dates]

        if any((d.month, d.day) != (1, 1) for d in dates):  # pragma: no cover
            warnings.warn('ignoring non -1-1 date(s) month/day: {!r}'.format(dates))

        start, end = dates
        return (
            -start.year if ma.group('start_sign') == '-' else start.year,
            -end.year if ma.group('end_sign') == '-' else end.year)

    @timespan.setter
    def timespan(self, value):
        if not (isinstance(value, (list, tuple)) and len(value) == 2):
            raise ValueError(value)

        # https://en.wikipedia.org/wiki/ISO_8601#Years
        if not all(-9999 <= v <= 9999 for v in value):
            warnings.warn('serializing year(s) outside the four-digit-range: {!r}'.format(value))

        def fmt(v):
            sign = '-' if v < 0 else ''
            return '{}{:04d}'.format(sign, abs(v))

        self._set('timespan', '{}-01-01/{}-01-01'.format(*map(fmt, value)))

    @property
    def links(self) -> typing.List[Link]:
        """
        Links to web resources related to the languoid
        """
        return [Link.from_string(s) for s in self.cfg.getlist(self.section_core, 'links')]

    @links.setter
    def links(self, value):
        assert isinstance(value, list)
        self._set('links', [v.to_string() for v in sorted(Link.from_(v) for v in value)])

    def update_links(self, domain, urls):
        new = [li for li in self.links if li.domain != domain] + [Link.from_(u) for u in urls]
        if set(new) != set(self.links):
            self.links = new
            return True
        return False

    @property
    def countries(self) -> typing.List[Country]:
        """
        Countries a language is spoken in.
        """
        return [Country.from_text(n)
                for n in self.cfg.getlist(self.section_core, 'countries')]

    @countries.setter
    def countries(self, value):
        assert isinstance(value, (list, tuple)) \
            and all(isinstance(o, Country) for o in value)
        self._set('countries', ['{0}'.format(c) for c in value])

    @property
    def name(self):
        """
        The Glottolog mame of the languoid
        """
        return self._get('name')

    @name.setter
    def name(self, value):
        self._set('name', value)

    @property
    def latitude(self) -> typing.Union[None, float]:
        """
        The geographic latitude of the point chosen as representative coordinate of the languoid
        """
        return self._get('latitude', float)

    @latitude.setter
    def latitude(self, value):
        self._set('latitude', round(float(value), 5))

    @property
    def longitude(self) -> typing.Union[None, float]:
        """
        The geographic longitude of the point chosen as representative coordinate of the languoid
        """
        return self._get('longitude', float)

    @longitude.setter
    def longitude(self, value):
        self._set('longitude', round(float(value), 5))

    @property
    def hid(self):
        return self._get('hid')

    @hid.setter
    def hid(self, value):
        self._set('hid', value)

    @property
    def level(self):
        if self._api:
            return self._get('level', self._api.languoid_levels.get)
        return self._get('level', lambda s: s)

    @level.setter
    def level(self, value):
        if self._api:
            self._set('level', self._api.languoid_levels.get(value).id)

    @property
    def iso(self):
        return self._get('iso639-3')

    @iso.setter
    def iso(self, value):
        self._set('iso639-3', value)

    @property
    def iso_code(self):
        return self._get('iso639-3')

    @iso_code.setter
    def iso_code(self, value):
        self._set('iso639-3', value)

    @property
    def iso_retirement(self):
        if 'iso_retirement' in self.cfg:
            kw = dict(self.cfg['iso_retirement'])
            if 'change_to' in kw:
                kw['change_to'] = self.cfg.getlist('iso_retirement', 'change_to')
            if 'comment' in kw:
                kw['comment'] = self.cfg.gettext('iso_retirement', 'comment')
            return ISORetirement(**kw)

    @property
    def fname(self):
        return self.dir.joinpath(INFO_FILENAME)



from subprocess import check_output
from collections import Counter

from clldutils.markup import Table

URL = 'https://github.com/lingpy/bibliography/raw/master/evobib-converted.bib'


def bibtool(bib):  # pragma: no cover
    args = ['bibtool', '-R']
    for k, v in [
        ('print.align.key', '0'),
        ('preserve.key.case', 'on'),
        ('print.line.length', '10000'),
        ('select.crossrefs', 'on'),
        ('expand.crossref', 'on'),
        ('new.entry.type', 'mvreference'),
        ('new.entry.type', 'mvbook'),
        ('new.entry.type', 'bookinbook'),
        ('new.entry.type', 'thesis'),
        ('new.entry.type', 'report'),
    ]:
        args.append('--{0}="{1}"'.format(k, v))
    args.append(bib.name)
    res = check_output(args, cwd=str(bib.parent))
    with bib.open(mode='wb') as fp:
        fp.write(res)


def update(newbib, bibfile, log):  # pragma: no cover
    bibtool(newbib)
    stats = Counter()

    def fix_entry_type(entry):
        type_ = entry.type.lower()
        type_map = {
            'thesis': 'phdthesis',
            'mvreference': 'misc',
            'mvbook': 'book',
            'bookinbook': 'book',
            'report': 'techreport',
        }
        entry.type = type_map.get(type_, type_)
        stats.update([entry.type])

    bibfile.update(newbib, log=log)
    bibfile.visit(fix_entry_type)
    bibfile.check(log)
    res = Table('entry type', '#')
    res.extend(list(stats.most_common()))
    res.append(['TOTAL', sum(stats.values())])
    print('\n' + res.render(tablefmt='simple'))



"""
List all metadata fields used in languoid INI files and their frequency.
"""
import collections

from clldutils.clilib import Table, add_format


def register(parser):
    add_format(parser)


def run(args):
    ops = collections.defaultdict(collections.Counter)

    for lang in args.repos.languoids():
        for secname, sec in lang.cfg.items():
            ops[secname].update(opt for opt, val in sec.items() if val)

    ops.pop('DEFAULT', None)

    with Table('section', 'option', 'count') as table:
        for section, options in ops.items():
            table.append([section, '', float(sum(options.values()))])
            for k, n in options.most_common():
                table.append(['', k, float(n)])



import os
import tempfile
import zipfile
from urllib.request import urlretrieve

URL = ('http://cdstar.shh.mpg.de/'
       '/bitstreams/EAEA0-A662-A618-386E-0/ldh_description.bib.zip')


def download(bibfile, log):  # pragma: no cover
    fname = 'description.bib'
    tmpdir = tempfile.gettempdir()
    zipped = os.path.join(tmpdir, fname + '.zip')
    urlretrieve(URL, zipped)
    zip = zipfile.ZipFile(zipped)
    zip.extract(fname, tmpdir)
    os.remove(zipped)

    bibfile.update(os.path.join(tmpdir, fname), log=log)
    bibfile.check(log)

    os.remove(os.path.join(tmpdir, fname))



"""
Display details of a Glottolog object.
"""
from pyglottolog.languoids import Reference
from pyglottolog.cli_util import get_languoid
from pyglottolog.util import sprint


def register(parser):
    parser.add_argument(
        'object',
        metavar='<GLOTTOCODE>|<ISO-CODE>|<BIBTEXKEY>',
    )


def run(args):
    if ':' in args.object:
        if args.object.startswith('**'):
            ref = Reference.from_string(args.object)
        else:
            ref = Reference(key=args.object)
        sprint('Glottolog reference {0}'.format(ref), attrs=['bold', 'underline'])
        print()
        src = ref.get_source(args.repos)
        sprint(src.text())
        print()
        sprint(src)
        return

    lang = get_languoid(args, args.object)
    print()
    sprint('Glottolog languoid {0}'.format(lang.id), attrs=['bold', 'underline'])
    print()
    sprint('Classification:', attrs=['bold', 'underline'])
    args.repos.ascii_tree(lang, maxlevel=1)
    print()
    sprint('Info:', attrs=['bold', 'underline'])
    sprint('Path: {0}'.format(lang.fname), 'green', attrs=['bold'])
    sources = lang.sources
    if sources:
        del lang.cfg['sources']['glottolog']
        del lang.cfg['sources']
    for line in lang.cfg.write_string().split('\n'):
        if not line.startswith('#'):
            sprint(line, None, attrs=['bold'] if line.startswith('[') else [])
    sprint('Sources:', attrs=['bold', 'underline'])
    for src in sources:
        src = src.get_source(args.repos)
        sprint(src.id, color='green')
        sprint(src.text())
        print()



"""
"""
from clldutils.clilib import PathType

from pyglottolog.references import evobib


def register(parser):
    parser.add_argument(
        'bib',
        type=PathType(type='file'),
        help='path to downloaded evobib-converted.bib')


def run(args):  # pragma: no cover
    evobib.update(args.bib, args.repos.bibfiles['evobib.bib'], args.log)
